﻿Public Class BinaryTree(Of T As IComparable(Of T))

    Private NotInheritable Class Node
        Public Property Item() As T
        Public Property Left() As Node
        Public Property Right() As Node
    End Class

    Public Enum TraverseModeEnum
        PreOrder
        PostOrder
        InOrder
        ReverseInOrder
    End Enum

    Public Property TraverseMode() As TraverseModeEnum

    Public Property Count() As Integer
        Get
            Return m_Count
        End Get
        Private Set(ByVal value As Integer)
            m_Count = value
        End Set
    End Property
    Private m_Count As Integer

    Private root As Node

    Public Sub Add(item As T)
        If root Is Nothing Then
            root = New Node() With {.Item = item}
        Else
            addTo(root, item)
        End If

        Count += 1
    End Sub

    Public Function Search(item As T) As T
        Dim node As Node = root

        While node IsNot Nothing
            Dim c As Integer = node.Item.CompareTo(item)

            If c = 0 Then
                Return node.Item
            End If

            If c > 0 Then
                node = node.Left
            Else
                node = node.Right
            End If
        End While
        Return Nothing
    End Function

    Public Overrides Function ToString() As String
        Dim s As String = ""
        Dim level As Integer = 0

        traverse(root, level, s)

        Return s
    End Function

    Private Sub traverse(node As Node, level As Integer, ByRef s As String)
        If node Is Nothing Then
            Return
        End If

        If TraverseMode = TraverseModeEnum.PreOrder Then
            s += "".PadLeft(level, " "c) + node.Item.ToString() + vbLf
            traverse(node.Left, level + 2, s)
            traverse(node.Right, level + 2, s)
        End If

        If TraverseMode = TraverseModeEnum.PostOrder Then
            traverse(node.Left, level + 2, s)
            traverse(node.Right, level + 2, s)
            s += "".PadLeft(level, " "c) + node.Item.ToString() + vbLf
        End If

        If TraverseMode = TraverseModeEnum.InOrder Then
            traverse(node.Left, level + 2, s)
            s += "".PadLeft(level, " "c) + node.Item.ToString() + vbLf
            traverse(node.Right, level + 2, s)
        End If
    End Sub

    Private Sub addTo(node As Node, item As T)
        If item.CompareTo(node.Item) < 0 Then
            If node.Left Is Nothing Then
                node.Left = New Node() With {.Item = item}
            Else
                addTo(node.Left, item)
            End If
        Else
            If node.Right Is Nothing Then
                node.Right = New Node() With {.Item = item}
            Else
                addTo(node.Right, item)
            End If
        End If
    End Sub


End Class
