﻿Public Class DoublyLinkedList(Of T)
    Implements IEnumerable(Of T)


    Private NotInheritable Class Node
        Public Property Item() As T
        Public Property Previous() As Node
        Public Property [Next]() As Node
    End Class

    Public Property Count() As Integer
        Get
            Return m_Count
        End Get
        Private Set
            m_Count = Value
        End Set
    End Property
    Private m_Count As Integer

    Public Overridable Sub Add(item As T)
        Dim newNode = New Node() With {
        .Item = item,
        .Previous = Nothing,
        .Next = Nothing
    }

        If first Is Nothing Then
            first = newNode
            last = newNode
        Else
            last.Next = newNode ' verketteter Knoten
            newNode.Previous = last
            last = newNode  ' neuer letzter Knoten
        End If
        Count += 1

    End Sub

    Public Function InsertAfter(previousItem As T, item As T) As Boolean
        Dim previousNode As Node = find(previousItem)
        If previousNode Is Nothing Then
            Return False
        End If

        Dim newNode As New Node() With {.Item = item}
        newNode.[Next] = previousNode.[Next]
        newNode.Previous = previousNode

        previousNode.[Next] = newNode

        If newNode.[Next] IsNot Nothing Then
            newNode.[Next].Previous = newNode
        Else
            last = newNode
        End If

        Count += 1

        Return True
    End Function


    Private first As Node, last As Node

    Default Public Property Item(index As Integer) As T
        Get
            Return findByIndex(index).Item
        End Get
        Set
            findByIndex(index).Item = Value
        End Set
    End Property

    Public Overridable Sub AddRange(items As T())
        For Each obj In items
            Add(obj)
        Next
    End Sub


    Public Function Contains(item As T) As Boolean
        Dim node = find(item)
        If node Is Nothing Then
            Return False
        End If
        Return True
    End Function

    Public Function Remove(item As T) As Boolean
        Dim node As Node = find(item)

        If node Is Nothing Then
            Return False
        End If

        ' Anfang löschen, first korrigieren
        If Object.ReferenceEquals(first, node) Then
            first = node.Next
        End If

        ' Ende löschen, last korrigieren
        If Object.ReferenceEquals(last, node) Then
            last = node.Previous
        End If

        ' Vorgängerverweis korrigieren
        If node.Previous IsNot Nothing Then
            node.Previous.Next = node.Next
        End If

        ' Nachfolgerverweis korrigieren
        If node.Next IsNot Nothing Then
            node.Next.Previous = node.Previous
        End If

        Count -= 1

        Return True
    End Function


    Public Sub Clear()
        first = Nothing
        last = Nothing
    End Sub

    Private Function Enumerator_GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        Return GetEnumerator()
    End Function

    Public Iterator Function GetEnumerator() As IEnumerator(Of T) Implements IEnumerable(Of T).GetEnumerator
        Dim node As Node = first

        While node IsNot Nothing
            Yield node.Item
            node = node.Next
        End While

        ' Return New LinkedListEnumerator(Me)
    End Function

    Private Function findPrevious(item As T) As Node
        Dim previousNode As Node = Nothing
        Dim node As Node = first

        While node IsNot Nothing
            If node.Item.Equals(item) Then
                Return previousNode
            End If

            previousNode = node
            node = node.Next
        End While
        Return Nothing
    End Function

    Private Function find(item As T) As Node
        Dim node As Node = first

        While node IsNot Nothing
            If node.Item.Equals(item) Then
                Return node
            End If

            node = node.Next
        End While
        Return Nothing
    End Function

    Private Function findByIndex(index As Integer) As Node
        If index >= Count Then
            Throw New IndexOutOfRangeException()
        End If

        Dim node As Node = first
        Dim i As Integer = 0

        While node IsNot Nothing
            If i = index Then
                Return node
            End If

            node = node.Next
            i += 1
        End While
        Return Nothing
    End Function

    Public Overrides Function ToString() As String
        Dim s As String = ""

        Dim node As Node = first

        While node IsNot Nothing
            s += node.Item.ToString() + " -> "
            node = node.[Next]
        End While
        s += "Count: " + Count.ToString()

        Return s
    End Function

End Class

