﻿Public Class HashtableDoubleHashing(Of TKey, TValue)
    Friend Class Pair(Of TPairKey, TPairValue)
        Public Property Key() As TPairKey
            Get
                Return m_Key
            End Get
            Private Set         ' key nicht nachträglich ändern
                m_Key = Value
            End Set
        End Property
        Private m_Key As TPairKey


        Public Property Value() As TPairValue
        Public Property IsDeleted() As Boolean

        Public Sub New(key As TPairKey, value As TPairValue, Optional isDeleted As Boolean = False)
            Me.Key = key
            Me.Value = value
            Me.IsDeleted = isDeleted
        End Sub
    End Class

    Private items As Pair(Of TKey, TValue)()

    Public Property Count() As Integer
        Get
            Return m_Count
        End Get
        Private Set
            m_Count = Value
        End Set
    End Property
    Private m_Count As Integer

    Public Sub New(Optional length As Integer = 0)
        length = If(length = 0, 500 * 2, length * 2)
        If length < 5 Then  ' siehe hash2
            length = 5
        End If

        length = calcPrimeLength(length)
        items = New Pair(Of TKey, TValue)(length - 1) {}
    End Sub

    Public Sub Add(key As TKey, value As TValue)
        If ContainsKey(key) Then
            Throw New ArgumentException("Item already exist in collection.")
        End If

        ' Überlauf, kein weiteres Element kann aufgenommen werden
        If Count + 1 = items.Length Then
            Throw New ArgumentException("No more space in array.")
        End If

        Dim hash As Integer = getHash(key)
        Dim steps As Integer = getHashSteps(key)

        While items(hash) IsNot Nothing AndAlso Not items(hash).Key.Equals(Nothing)
            hash += steps
            ' Arrayüberlauf verhindern
            hash = hash Mod items.Length
        End While

        items(hash) = New Pair(Of TKey, TValue)(key, value)

        Count += 1
    End Sub

    Public Sub Remove(key As TKey)
        Dim hash As Integer = getHash(key)
        Dim steps As Integer = getHashSteps(key)

        While items(hash) IsNot Nothing
            If items(hash).Key.Equals(key) AndAlso Not items(hash).IsDeleted Then
                ' items[hash] nicht auf null setzen, da sonst beim Suchen dahinterliegende nicht mehr gefunden werden
                ' Kennzeichen, das Eintrag gelöscht. Kein Umkopieren im Array, da sich sonst der Index verändert und der Hash nicht 
                ' mehr passt
                items(hash).IsDeleted = True
                Count -= 1
                Exit While
            End If
            hash += steps
            ' Arrayüberlauf verhindern
            hash = hash Mod items.Length
        End While
    End Sub

    Default Public Property Item(key As TKey) As TValue
        Get
            Dim hash As Integer = getHash(key)
            Dim steps As Integer = getHashSteps(key)

            While items(hash) IsNot Nothing
                If items(hash).Key.Equals(key) Then
                    Return items(hash).Value
                End If
                hash += steps
                ' Arrayüberlauf verhindern
                hash = hash Mod items.Length
            End While
            Throw New ArgumentException("Key not found.")
        End Get
        Set
            If Not ContainsKey(key) Then
                Add(key, Value)
            Else
                update(key, Value)
            End If
        End Set
    End Property

    Public Sub Clear()
        items = New Pair(Of TKey, TValue)(items.Length - 1) {}
        Count = 0
    End Sub

    Public Function Contains(key As TKey) As Boolean
        Return ContainsKey(key)
    End Function

    Public Function ContainsKey(key As TKey) As Boolean
        Dim hash As Integer = getHash(key)
        Dim steps As Integer = getHashSteps(key)

        While items(hash) IsNot Nothing
            If items(hash).Key.Equals(key) AndAlso Not items(hash).IsDeleted Then
                Return True
            End If
            hash += steps
            ' Arrayüberlauf verhindern
            hash = hash Mod items.Length
        End While
        Return False
    End Function

    Public Overrides Function ToString() As String
        Dim s As String = ""

        For i As Integer = 0 To items.Length - 1
            If items(i) IsNot Nothing AndAlso Not items(i).Key.Equals(Nothing) Then
                s += items(i).Key.ToString() + "|" + items(i).Value.ToString() + " -> "
            End If
        Next
        Return (s & Convert.ToString("Count: ")) + Count.ToString()
    End Function

    Private Sub update(key As TKey, value As TValue)
        Dim hash As Integer = getHash(key)
        Dim steps As Integer = getHashSteps(key)

        While items(hash) IsNot Nothing AndAlso Not items(hash).IsDeleted
            If items(hash).Key.Equals(key) Then
                items(hash).Value = value
                Exit While
            End If
            hash += steps
            ' Arrayüberlauf verhindern
            hash = hash Mod items.Length
        End While
    End Sub

    Private Function calcPrimeLength(length As Integer) As Integer
        While Not isPrime(length)
            length += 1
        End While

        Return length
    End Function

    Private Function isPrime(number As Integer) As Boolean
        For i As Integer = 2 To number / 2
            If number Mod i = 0 Then
                Return False
            End If
        Next
        Return True
    End Function

    Private Function getHash(key As TKey) As Integer
        Return Math.Abs(key.GetHashCode()) Mod items.Length
    End Function

    Private Function getHashSteps(key As TKey) As Integer
        Return 5 - (key.GetHashCode() Mod 5)
    End Function

End Class
