﻿Public Class HashtableLinearProbing(Of TKey, TValue)
    Friend Class Pair(Of TPairKey, TPairValue)
        Public Property Key() As TPairKey
            Get
                Return m_Key
            End Get
            Private Set
                m_Key = Value
            End Set
        End Property
        Private m_Key As TPairKey

        Public Property Value() As TPairValue
        Public Property IsDeleted() As Boolean

        Public Sub New(key As TPairKey, value As TPairValue)
            Me.Key = key
            Me.Value = value
        End Sub
    End Class

    Private items As Pair(Of TKey, TValue)()

    Public Property Count() As Integer
        Get
            Return m_Count
        End Get
        Private Set
            m_Count = Value
        End Set
    End Property
    Private m_Count As Integer

    Default Public ReadOnly Property Item(key As TKey) As TValue
        Get
            Dim hash As Integer = Math.Abs(key.GetHashCode()) Mod items.Length

            While items(hash) IsNot Nothing
                If items(hash).Key.Equals(key) AndAlso Not items(hash).IsDeleted Then
                    Return items(hash).Value
                End If
                hash += 1
                hash = hash Mod items.Length
            End While
            Throw New ArgumentException("Key not found.")
        End Get
    End Property

    Public Sub New(Optional length As Integer = 0)
        length = calcPrimeLength(If(length = 0, 500 * 2, length * 2))
        items = New Pair(Of TKey, TValue)(length - 1) {}
    End Sub

    Public Sub Add(key As TKey, value As TValue)
        If ContainsKey(key) Then
            Throw New ArgumentException("Item already exist.")
        End If

        ' Überlauf, kein weiteres Element kann aufgenommen werden
        If Count + 1 = items.Length Then
            Throw New ArgumentException("No more space in array.")
        End If

        Dim hash As Integer = Math.Abs(key.GetHashCode()) Mod items.Length

        While items(hash) IsNot Nothing AndAlso Not items(hash).IsDeleted
            hash += 1
            hash = hash Mod items.Length
        End While

        items(hash) = New Pair(Of TKey, TValue)(key, value)

        Count += 1
    End Sub

    Public Sub Remove(key As TKey)
        Dim hash As Integer = Math.Abs(key.GetHashCode()) Mod items.Length

        While items(hash) IsNot Nothing
            If items(hash).Key.Equals(key) AndAlso Not items(hash).IsDeleted Then
                items(hash).IsDeleted = True
                Count -= 1
                Exit While
            End If
            hash += 1
            hash = hash Mod items.Length
        End While
    End Sub

    Public Function ContainsKey(key As TKey) As Boolean
        Dim hash As Integer = Math.Abs(key.GetHashCode()) Mod items.Length

        While items(hash) IsNot Nothing AndAlso Not items(hash).IsDeleted
            If items(hash).Key.Equals(key) Then
                Return True
            End If
            hash += 1
            hash = hash Mod items.Length
        End While
        Return False
    End Function

    Public Overrides Function ToString() As String
        Dim s As String = ""

        For i As Integer = 0 To items.Length - 1
            If items(i) IsNot Nothing AndAlso Not items(i).IsDeleted Then
                s += items(i).Key.ToString() + "|" + items(i).Value.ToString() + " -> "
            End If
        Next
        Return (s & Convert.ToString("Count: ")) + Count.ToString()
    End Function

    Private Function calcPrimeLength(length As Integer) As Integer
        While Not isPrime(length)
            length += 1
        End While

        Return length
    End Function

    Private Function isPrime(number As Integer) As Boolean
        For i As Integer = 2 To number / 2
            If number Mod i = 0 Then
                Return False
            End If
        Next
        Return True
    End Function
End Class
