﻿Imports System.IO
Imports System.Net
Imports System.Net.Sockets

Public Class HttpServer
    Public Property Path As String
    Public Property Port As Integer
    Public Property listener As TcpListener

    Public Sub New(port As Integer, path As String)
        Me.Port = port
        Me.Path = path
    End Sub

    Public Sub Start()
        Dim requests = New Requests()

        Dim address As IPAddress = IPAddress.Parse("127.0.0.1")

        listener = New TcpListener(address, Port)

        listener.Start()

        While True
            Dim c = listener.AcceptTcpClient()

            requests.Add(c)

            If Not listener.Pending() Then
                requests.Process(Path)
            End If
        End While

    End Sub
End Class

Friend Class Requests
    Private queue As New Queue(Of TcpClient)()

    Public Sub Add(c As TcpClient)
        queue.Enqueue(c)
    End Sub

    Public Sub Process(path As String)
        Console.WriteLine("Queue: " + queue.Count.ToString())

        While queue.Count > 0
            Try
                Dim r = New Request(queue.Dequeue(), path)
                Console.WriteLine("  " + r.LocalUrl)
            Catch e As Exception
                Console.WriteLine("Error: " + e.Message)
            End Try
        End While
    End Sub


End Class

Friend Class Request
    Private c As TcpClient
    Private path As String

    Public Property LocalUrl() As String
        Get
            Return m_LocalUrl
        End Get
        Private Set
            m_LocalUrl = Value
        End Set
    End Property
    Private m_LocalUrl As String

    Public Property Verb() As String
        Get
            Return m_Verb
        End Get
        Private Set
            m_Verb = Value
        End Set
    End Property
    Private m_Verb As String

    Public Property HttpVersion() As String
        Get
            Return m_HttpVersion
        End Get
        Private Set
            m_HttpVersion = Value
        End Set
    End Property
    Private m_HttpVersion As String

    Public Property Header() As List(Of String)
        Get
            Return m_Header
        End Get
        Private Set
            m_Header = Value
        End Set
    End Property
    Private m_Header As List(Of String)

    Public Sub New(c As TcpClient, path As String)
        Me.c = c
        Me.path = path

        Header = New List(Of String)()

        process()
    End Sub

    Private Sub process()
        Dim r = New StreamReader(c.GetStream())
        Dim w = New StreamWriter(c.GetStream())

        Try
            handleRequest(r, w)
        Catch e As Exception
            failure(w)
            Throw e
        Finally
            w.Close()
            r.Close()
            c.Close()
        End Try
    End Sub

    Private Sub success(stream As StreamWriter, ext As String)
        stream.WriteLine("HTTP/1.0 200 OK")
        stream.WriteLine(Convert.ToString("Content-Type: text/") & ext)
        stream.WriteLine("Connection: close")
        stream.WriteLine("")
    End Sub

    Private Sub failure(stream As StreamWriter)
        stream.WriteLine("HTTP/1.0 404 not found")
        stream.WriteLine("Connection: close")
        stream.WriteLine("")
    End Sub

    Private Sub readHeader(r As StreamReader)
        Dim attribute As String = r.ReadLine()

        While attribute IsNot Nothing
            If attribute.Equals("") Then
                Exit While
            End If
            Header.Add(attribute)
            attribute = r.ReadLine()
        End While
    End Sub

    Private Sub handleRequest(r As StreamReader, w As StreamWriter)
        Dim request As String = r.ReadLine()

        If request IsNot Nothing Then
            Dim tokens As String() = request.Split(" "c)

            Verb = tokens(0).ToUpper()
            LocalUrl = tokens(1)
            HttpVersion = tokens(2)

            readHeader(r)

            Dim f As String = System.IO.Path.Combine(System.IO.Path.GetFullPath(path), LocalUrl.TrimStart("/"c))

            Dim text As String = File.ReadAllText(f)

            success(w, System.IO.Path.GetExtension(LocalUrl).TrimStart("."c))

            w.Write(text)
        End If
    End Sub


End Class
